# Code to support running an ast at a remote func-adl server.
import ast
import logging
from abc import ABC
from collections.abc import Iterable
from typing import Any, Optional, TypeVar, Union, cast

from func_adl import EventDataset
from qastle import python_ast_to_text_ast
from servicex import ServiceXDataset
from servicex.utils import DatasetType


class FuncADLServerException (Exception):
    'Thrown when an exception happens contacting the server'
    def __init__(self, msg):
        Exception.__init__(self, msg)


T = TypeVar('T')


class ServiceXDatasetSourceBase(EventDataset[T], ABC):
    '''
    Base class for a ServiceX backend dataset.

    While no methods are abstract, base classes will need to add arguments
    to the base `EventDataset` to make sure that it contains the information
    backends expect!
    '''
    # How we map from func_adl to a servicex query
    _ds_map = {
        'ResultTTree': 'get_data_rootfiles_async',
        'ResultParquet': 'get_data_parquet_async',
        'ResultPandasDF': 'get_data_pandas_df_async',
        'ResultAwkwardArray': 'get_data_awkward_async',
    }

    # If it comes down to format, what are we going to grab?
    _format_map = {
        'root': 'get_data_rootfiles_async',
        'parquet': 'get_data_parquet_async',
    }

    # These are methods that are translated locally
    _execute_locally = ['ResultPandasDF', 'ResultAwkwardArray']

    # If we have a choice of formats, what can we do, in
    # prioritized order?
    _format_list = ['parquet', 'root']

    def __init__(self, sx: Union[ServiceXDataset, DatasetType], backend_name: str, item_type: type = Any):
        '''
        Create a servicex dataset sequence from a servicex dataset
        '''
        super().__init__(item_type=item_type)

        # Get the base created
        if isinstance(sx, (str, Iterable)):
            ds = ServiceXDataset(sx, backend_name=backend_name)
        else:
            ds = sx
        self._ds = ds

        self._return_qastle = False

    @property
    def return_qastle(self) -> bool:
        '''Get/Set `qastle` generation flag.

        If `True`, then execution of this query will return `qastle`, and if `False` then
        the query will be executed.
        '''
        return self._return_qastle

    @return_qastle.setter
    def return_qastle(self, value: bool):
        self._return_qastle = value

    def check_data_format_request(self, f_name: str):
        '''Check to make sure the dataformat that is getting requested is ok. Throw an error
        to give the user enough undersanding of why it isn't.

        Args:
            f_name (str): The function name of the final thing we are requesting.
        '''
        if ((f_name == 'ResultTTree' and self._ds.first_supported_datatype('root') is None)
                or (f_name == 'ResultParquet'
                    and self._ds.first_supported_datatype('parquet') is None)):
            raise FuncADLServerException(f'{f_name} is not supported by {str(self._ds)}')

        # If here, we assume the user knows what they are doing. The return format will be
        # the default file type
        return

    def generate_qastle(self, a: ast.Call) -> str:
        '''Generate the qastle from the ast of the query.

        1. The top level function is already marked as being "ok"
        1. If the top level function is something we have to process locally,
           then we strip it off.

        Args:
            a (ast.AST): The complete AST of the request.

        Returns:
            str: Qastle that should be sent to servicex
        '''
        top_function = cast(ast.Name, a.func).id
        source = a
        if top_function in self._execute_locally:
            # Request the default type here
            default_format = self._ds.first_supported_datatype(['parquet', 'root'])
            assert default_format is not None, 'Unsupported ServiceX returned format'
            method_to_call = self._format_map[default_format]

            stream = a.args[0]
            col_names = a.args[1]
            if method_to_call == 'get_data_rootfiles_async':
                # If we have no column names, then we must be using a dictionary to set them - so just pass that
                # directly.
                assert isinstance(col_names, (ast.List, ast.Constant, ast.Str)), f'Programming error - type name not known {type(col_names).__name__}'
                if isinstance(col_names, ast.List) and len(col_names.elts) == 0:
                    source = stream
                else:
                    source = ast.Call(
                        func=ast.Name(id='ResultTTree', ctx=ast.Load()),
                        args=[stream, col_names, ast.Str('treeme'), ast.Str('junk.root')])
            elif method_to_call == 'get_data_parquet_async':
                source = stream
                # See #32 for why this is commented out
                # source = ast.Call(
                #     func=ast.Name(id='ResultParquet', ctx=ast.Load()),
                #     args=[stream, col_names, ast.Str('junk.parquet')])
            else:  # pragma: no cover
                # This indicates a programming error
                assert False, f'Do not know how to call {method_to_call}'

        return python_ast_to_text_ast(source)

    async def execute_result_async(self, a: ast.Call, title: Optional[str] = None) -> Any:
        r'''
        Run a query against a func-adl ServiceX backend. The appropriate part of the AST is
        shipped there, and it is interpreted.

        Arguments:

            a:                  The ast that we should evaluate
            title:              Optional title to be added to the transform

        Returns:
            v                   Whatever the data that is requested (awkward arrays, etc.)
        '''
        # Check the call is legal for this datasource.
        a_func = cast(ast.Name, a.func)
        self.check_data_format_request(a_func.id)

        # Get the qastle string for this query
        q_str = self.generate_qastle(a)
        logging.getLogger(__name__).debug(f'Qastle string sent to servicex: {q_str}')

        # If only qastle is wanted, return here.
        if self.return_qastle:
            return q_str

        # Find the function we need to run against.
        if a_func.id in self._ds_map:
            name = self._ds_map[a_func.id]
        else:
            data_type = self._ds.first_supported_datatype(['parquet', 'root'])
            if data_type is not None and data_type in self._format_map:
                name = self._format_map[data_type]
            else:
                raise FuncADLServerException(f'Internal error - asked for {a_func.id} - but this dataset does not support it.')

        # Run ghe query for real!
        attr = getattr(self._ds, name)
        return await attr(q_str, title=title)


class ServiceXSourceCPPBase(ServiceXDatasetSourceBase[T]):
    def __init__(self, sx: Union[ServiceXDataset, DatasetType], backend_name: str, item_type: type = Any):
        '''Create a C++ backend data set source

        Args:
            sx (Union[ServiceXDataset, str]): The ServiceX dataset or dataset source.
            backend_name (str): The backend type, `xaod`, for example, for the ATLAS R21 xaod
        '''
        super().__init__(sx, backend_name, item_type)

        # Add the filename
        self.query_ast.args.append(ast.Str(s='bogus.root'))  # type: ignore


class ServiceXSourceXAOD(ServiceXSourceCPPBase[T]):
    def __init__(self, sx: Union[ServiceXDataset, DatasetType], backend='xaod', item_type: type = Any):
        '''
        Create a servicex dataset sequence from a servicex dataset
        '''
        super().__init__(sx, backend, item_type)


class ServiceXSourceCMSRun1AOD(ServiceXSourceCPPBase[T]):
    def __init__(self, sx: Union[ServiceXDataset, DatasetType], backend='cms_run1_aod', item_type: type = Any):
        '''
        Create a servicex dataset sequence from a servicex dataset
        '''
        super().__init__(sx, backend, item_type)


class ServiceXSourceUpROOT(ServiceXDatasetSourceBase[T]):
    def __init__(self, sx: Union[ServiceXDataset, DatasetType], treename: str, backend_name='uproot', item_type: type = Any):
        '''
        Create a servicex dataset sequence from a servicex dataset
        '''
        super().__init__(sx, backend_name, item_type)

        # Modify the argument list in EventDataSset to include a dummy filename and
        # tree name
        self.query_ast.args.append(ast.Str(s='bogus.root'))  # type: ignore
        self.query_ast.args.append(ast.Str(s=treename))  # type: ignore



from abc import ABC, abstractmethod
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

import aiohttp
from func_adl_xAOD.atlas.xaod import xAODDataset
from func_adl_xAOD.cms.aod import CMSRun1AODDataset
from func_adl_xAOD.common.local_dataset import LocalDataset
from qastle import text_ast_to_python_ast
from servicex import ServiceXDataset
from servicex.utils import ServiceXUnknownDataRequestID

from .ServiceX import ServiceXSourceCPPBase


class _local_file_copier:
    def __init__(self):
        self._request_info_dict: Dict[str, List[Path]] = {}

    def associate_file(self, request_id: str, files: List[Path]):
        '''These are the files that should be returned when the query

        Args:
            request_id (str): The request id
            files (List[str]): The files to attach to this request id
        '''
        self._request_info_dict[request_id] = files

    def get_files(self, request_id: str) -> List[str]:
        '''When the run has finished, return the file.

        Args:
            request_id (str): The request ID we are doing this for

        Returns:
            List[str]: List of the files
        '''
        if request_id not in self._request_info_dict:
            raise ServiceXUnknownDataRequestID(f'Files for request {request_id} not found')

        return [f.name for f in self._request_info_dict[request_id]]

    async def download_file(self,
                            request_id: str,
                            bucket_name: str,
                            output_file: Path) -> None:
        'Copy the files from the request id to the appropriate output location'
        # Find the proper file to copy over from our list of files.
        all_files = [f for f in self._request_info_dict[request_id] if f.name == bucket_name]
        all_files[0].rename(output_file)


class _sx_local_file_minio_factory:
    def __init__(self):
        self._local = _local_file_copier()

    def associate_file(self, request_id: str, files: List[Path]):
        'Pass the association down the line'
        self._local.associate_file(request_id, files)

    def from_best(self, _: Optional[Dict[str, str]] = None):
        'Since this is never "complex", we can just return a one-off'
        return self._local


class _sx_local_file_adaptor:
    def __init__(self, ds: LocalDataset, minio: _sx_local_file_minio_factory):
        '''Create a ServiceXAdaptor that will run the local file accessor.
        '''
        self._ds = ds
        self._minio = minio

    async def submit_query(self, _: aiohttp.ClientSession,
                           json_query: Dict[str, str]) -> Dict[str, str]:
        '''Kick off the docker run, and then return, right away, a response
        that will mock the ServiceX backend.

        Args:
            _ (aiohttp.ClientSession): The http context (ignored)
            json_query (Dict[str, str]): Contains the info on the query

        Returns:
            Dict[str, str]: The reply from ServiceX
        '''
        # Generate a new unique ID for this query
        request_id = str(uuid.uuid4())

        # Next, run the thing, and wait for it to finish
        query = json_query['selection']
        title = json_query['title'] if 'title' in json_query else ''
        file_list = await self._ds.execute_result_async(
            text_ast_to_python_ast(query).body[0].value,
            title
        )

        self._minio.associate_file(request_id, file_list)

        # Return the info needed from the submit!
        return {
            'request_id': request_id,
        }

    async def get_query_status(self, _: aiohttp.ClientSession,
                               request_id: str) -> Dict[str, str]:
        '''Returns the full query information from the endpoint.
        Args:
            client (aiohttp.ClientSession): Client session on which to make the request.
            request_id (str): The request id to return the tranform status
        Raises:
            ServiceXException: If we fail to find the information.
        Returns:
            Dict[str, str]: The JSON dictionary of information returned from ServiceX
        '''
        return {
            'request_id': request_id,
            'files-remaining': '0',
            'files-skipped': '0',
            'files-processed': '1',
        }

    async def get_transform_status(self, client: aiohttp.ClientSession, request_id: str) -> \
            Tuple[Optional[int], int, Optional[int]]:
        return (0, 1, 0)


T = TypeVar('T')


class SXLocalCPP(ServiceXSourceCPPBase[T], ABC):
    def __init__(self, files: Union[str, List[str], Path, List[Path]],
                 docker_image: Optional[str] = None,
                 docker_tag: Optional[str] = None,
                 item_type: type = Any):
        '''A local ServiceX-like dataset. Used for the C++ backends

        NOTE: This version of is not suitable for running large numbers of files or
        datasets simultantiously on your local machine! Everything is run serially.

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
        '''

        # Create the local dataset
        ds = self._create_dataset(files, docker_image, docker_tag)

        # To create the ServiceXDataset
        minio_local = _sx_local_file_minio_factory()
        sx_ds = ServiceXDataset(
            'local_dataset',
            backend_name=self._get_backend_type(),
            servicex_adaptor=_sx_local_file_adaptor(ds, minio_local),  # type: ignore
            minio_adaptor=minio_local,  # type: ignore
            status_callback_factory=None,
        )

        # And create our home body!
        super().__init__(sx_ds, self._get_backend_type(), item_type)

    @classmethod
    @abstractmethod
    def _create_dataset(cls, files: Union[str, List[str], Path, List[Path]],
                        docker_image: Optional[str] = None,
                        docker_tag: Optional[str] = None) -> LocalDataset:
        '''Create a dataset that can be used with the C++ backends

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
            docker_image (Optional[str]): The docker image to use.
            docker_tag (Optional[str]): The docker tag to use.
        '''

    @classmethod
    @abstractmethod
    def _get_backend_type(cls) -> str:
        'Return the backend type/name string'


U = TypeVar('U')


class SXLocalxAOD(SXLocalCPP[U]):
    @classmethod
    def _create_dataset(cls, files: Union[str, List[str], Path, List[Path]],
                        docker_image: Optional[str] = None,
                        docker_tag: Optional[str] = None) -> LocalDataset:
        '''Create a dataset that can be used with the C++ backends

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
            docker_image (Optional[str]): The docker image to use.
            docker_tag (Optional[str]): The docker tag to use.
        '''
        extra_args = {}
        if docker_image is not None:
            extra_args['docker_image'] = docker_image
        if docker_tag is not None:
            extra_args['docker_tag'] = docker_tag
        return xAODDataset(files, **extra_args)

    @classmethod
    def _get_backend_type(cls) -> str:
        'Return code for a xaod backend'
        return 'xaod'

    def __init__(self, files: Union[str, List[str], Path, List[Path]],
                 docker_image: Optional[str] = None,
                 docker_tag: Optional[str] = None,
                 item_type: type = Any):
        '''A local ServiceX-like dataset. Will run locally, in docker, synchronously.

        NOTE: This version of is not suitable for running large numbers of files or
        datasets simultantiously on your local machine! Everything is run serially.

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
        '''
        super().__init__(files, docker_image, docker_tag, item_type)


V = TypeVar('V')


class SXLocalCMSRun1AOD(SXLocalCPP[V]):
    @classmethod
    def _create_dataset(cls, files: Union[str, List[str], Path, List[Path]],
                        docker_image: Optional[str] = None,
                        docker_tag: Optional[str] = None) -> LocalDataset:
        '''Create a dataset that can be used with the CMSRun1 backend

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
            docker_image (Optional[str]): The docker image to use.
            docker_tag (Optional[str]): The docker tag to use.
        '''
        extra_args = {}
        if docker_image is not None:
            extra_args['docker_image'] = docker_image
        if docker_tag is not None:
            extra_args['docker_tag'] = docker_tag
        return CMSRun1AODDataset(files, **extra_args)

    @classmethod
    def _get_backend_type(cls) -> str:
        'Return code for a cms_run1_aod backend'
        return 'cms_run1_aod'

    def __init__(self, files: Union[str, List[str], Path, List[Path]],
                 docker_image: Optional[str] = None,
                 docker_tag: Optional[str] = None,
                 item_type: type = Any):
        '''A local ServiceX-like dataset. Will run locally, in docker, synchronously.

        NOTE: This version of is not suitable for running large numbers of files or
        datasets simultantiously on your local machine! Everything is run serially.

        Args:
            files (Union[str, List[str], Path, List[Path]]): List of files to run on.
        '''
        super().__init__(files, docker_image, docker_tag, item_type)



# Main two things from this package
# flake8: noqa
from .ServiceX import ServiceXSourceUpROOT, ServiceXSourceXAOD, ServiceXSourceCMSRun1AOD, FuncADLServerException  # NOQA
try:
    from .local_dataset import SXLocalxAOD, SXLocalCMSRun1AOD  # NOQA
except ImportError:
    pass
